# Utility file to generate a certificate for the purposes of testing the KMIP mgmt apis

from datetime import datetime, timedelta


def generate_selfsigned_cert(hostname, key=None):
    """Generates self signed certificate for a hostname, and optional IP addresses."""
    from cryptography import x509
    from cryptography.x509.oid import NameOID
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric import rsa

    # Generate our key
    if key is None:
        key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
            backend=default_backend(),
        )

    name = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, hostname)])

    # best practice seem to be to include the hostname in the SAN, which *SHOULD* mean COMMON_NAME is ignored.
    alt_names = [x509.DNSName(hostname)]

    san = x509.SubjectAlternativeName(alt_names)

    # path_len=0 means this cert can only sign itself, not other certs.
    basic_contraints = x509.BasicConstraints(ca=True, path_length=0)
    now = datetime.utcnow()
    cert = (
        x509.CertificateBuilder()
        .subject_name(name)
        .issuer_name(name)
        .public_key(key.public_key())
        .serial_number(1000)
        .not_valid_before(now)
        .not_valid_after(now + timedelta(days=10 * 365))
        .add_extension(basic_contraints, False)
        .add_extension(san, False)
        .sign(key, hashes.SHA256(), default_backend())
    )
    cert_pem = cert.public_bytes(encoding=serialization.Encoding.PEM)

    return cert_pem.decode("utf-8")



import logging
import os
import test.integration.self_signed_cert as self_signed_cert
import time
import unittest

import redstone


class KeyProtectTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.rc = redstone.service("ResourceController")
        cls.instance_id, cls.crn = cls.rc.create_instance(
            name="redstone-keyprotect-integration-tests",
            plan_id=cls.rc.KEYPROTECT_PLAN_ID,
            region="us-south",
        )

        cls.kp = redstone.service(
            "KeyProtect", region="us-south", service_instance_id=cls.instance_id
        )

    @classmethod
    def tearDownClass(cls):
        try:
            for key in cls.kp.keys():
                cls.kp.delete(key.get("id"))
        finally:
            cls.rc.delete_instance(cls.instance_id)

    def test_create_key(self):
        # create a key
        resp = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, resp.get("id"))
        self.assertEqual(resp["state"], 1)
        self.assertEqual(resp["name"], "test-key")

    def test_create_key_with_alias(self):
        # create a key with aliases
        resp = self.kp.create(
            name="test-key",
            root=True,
            alias_list=["key_alias_1", "key_alias_2", "key_alias_3"],
        )
        self.addCleanup(self.kp.delete, resp.get("id"))
        self.assertEqual(resp["state"], 1)
        self.assertEqual(resp["name"], "test-key")
        self.assertEqual(len(resp["aliases"]), 3)
        self.assertEqual(resp["aliases"][0], "key_alias_1")

    def test_create_key_with_more_then_5_alias(self):
        # create a key with more than 5 aliases, should raise ValueError
        with self.assertRaises(ValueError):
            self.kp.create(
                name="test-key",
                root=True,
                alias_list=[
                    "key_alias_1",
                    "key_alias_2",
                    "key_alias_3",
                    "key_alias_4",
                    "key_alias_5",
                    "key_alias_6",
                ],
            )

    def test_get_key(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))
        resp = self.kp.get_key(key_id_or_alias=key.get("id"))
        self.assertEqual(resp["state"], 1)
        self.assertEqual(resp["name"], "test-key")

    def test_get_key_using_alias(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True, alias_list=["key_alias"])
        self.addCleanup(self.kp.delete, key.get("id"))
        resp = self.kp.get_key(key_id_or_alias="key_alias")
        self.assertEqual(resp["state"], 1)
        self.assertEqual(resp["name"], "test-key")
        self.assertEqual(resp["aliases"][0], "key_alias")

    def test_wrap_unwrap(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))
        # wrap
        message = b"This is a really important message."
        wrapped = self.kp.wrap(key.get("id"), message)
        ciphertext = wrapped.get("ciphertext")
        # unwrap
        unwrapped = self.kp.unwrap(key.get("id"), ciphertext)
        self.assertEqual(message, unwrapped)

    def test_wrap_unwrap_with_aad(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))
        # wrap
        message = b"This is a really important message."
        wrapped = self.kp.wrap(key.get("id"), message, aad=["python-keyprotect"])
        ciphertext = wrapped.get("ciphertext")
        # unwrap
        unwrapped = self.kp.unwrap(key.get("id"), ciphertext, aad=["python-keyprotect"])
        self.assertEqual(message, unwrapped)

    def test_wrap_rewrap_unwrap(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))
        # wrap
        message = b"This is a really important message."
        wrapped = self.kp.wrap(key.get("id"), message)
        ciphertext = wrapped.get("ciphertext")
        # rotate
        self.kp.rotate_key(key.get("id"))
        # rewrap
        rewrapped = self.kp.rewrap(key.get("id"), ciphertext)
        ciphertext = rewrapped.get("ciphertext")
        # unwrap
        unwrapped = self.kp.unwrap(key.get("id"), ciphertext)
        self.assertEqual(message, unwrapped)

    def test_disable_enable_key(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))

        # disable
        self.kp.disable_key(key.get("id"))
        resp = self.kp.get(key.get("id"))
        self.assertEqual(resp["state"], 2)

        # enable, must wait 30 sec or more before sending an enable
        time.sleep(30)

        self.kp.enable_key(key.get("id"))
        resp = self.kp.get(key.get("id"))
        self.assertEqual(resp["state"], 1)

    def test_restore_key(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        key_id = key.get("id")
        self.addCleanup(self.kp.delete, key_id)

        # delete the key
        self.kp.delete(key_id)
        resp = self.kp.get(key_id)
        # state Destroyed = 5
        self.assertEqual(resp["state"], 5)

        # restore, must wait 30 sec or more before sending an restore
        time.sleep(30)

        self.kp.restore_key(key_id)
        resp = self.kp.get(key_id)
        # state Active = 1
        self.assertEqual(resp["state"], 1)

    def test_key_ring(self):
        # test creating key ring
        resp = self.kp.create_key_ring("testKeyRingIdPython")

        # test getting list of key ring
        resp = self.kp.get_key_rings()
        for resource in resp["resources"]:
            if resource["id"] != "default":
                self.assertEqual(resource["id"], "testKeyRingIdPython")

        # test deleting key ring
        resp = self.kp.delete_key_ring("testKeyRingIdPython")

    def test_import_token(self):
        # create import token
        resp = self.kp.create_import_token(expiration=86400, max_allowed_retrievals=50)
        self.assertEqual(resp["maxAllowedRetrievals"], 50)
        self.assertEqual(resp["remainingRetrievals"], 50)

        # get import token
        resp = self.kp.get_import_token()
        self.assertEqual(resp["maxAllowedRetrievals"], 50)
        self.assertEqual(resp["remainingRetrievals"], 49)

    def test_key_alias(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))

        # test creating key alias
        resp = self.kp.create_key_alias(key["id"], "testKeyAlias")
        self.assertEqual(resp["resources"][0]["keyId"], key["id"])
        self.assertEqual(resp["resources"][0]["alias"], "testKeyAlias")

        # test getting key by using the key alias
        resp = self.kp.get("testKeyAlias")
        self.assertEqual(resp["id"], key["id"])
        self.assertEqual(resp["aliases"][0], "testKeyAlias")

        # test deleting key alias
        resp = self.kp.delete_key_alias(key["id"], "testKeyAlias")
        self.assertEqual(resp, None)

    def test_get_registrations(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))

        # get registrations associated with a key
        resp1 = self.kp.get_registrations(key["id"])
        self.assertEqual(resp1["metadata"]["collectionTotal"], 0)

        #  get registrations associated with an instance
        resp2 = self.kp.get_registrations()
        self.assertGreaterEqual(
            resp2["metadata"]["collectionTotal"], resp1["metadata"]["collectionTotal"]
        )

    def test_key_policies(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        self.addCleanup(self.kp.delete, key.get("id"))

        # test key rotation policy
        resp = self.kp.set_key_rotation_policy(key["id"], rotation_interval=2)
        self.assertEqual(resp["resources"][0]["rotation"]["interval_month"], 2)

        # test key dual auth delete policy, set to False so that it can be cleaned up
        resp = self.kp.set_key_dual_auth_policy(key["id"], dual_auth_enable=False)
        self.assertFalse(resp["resources"][0]["dualAuthDelete"]["enabled"])

        # test get key policies
        resp = self.kp.get_key_policies(key["id"])

        self.assertEqual(len(resp["resources"]), 2)
        for resource in resp["resources"]:
            if "rotation" in resource:
                self.assertEqual(resource["rotation"]["interval_month"], 2)
            else:
                self.assertFalse(resource["dualAuthDelete"]["enabled"])

    def test_instance_policies(self):
        # set instance dual auth delete policy
        self.kp.set_instance_dual_auth_policy(dual_auth_enable=False)

        # set instance allowed network policy
        self.kp.set_instance_allowed_network_policy(
            allowed_network_enable=True, network_type="public-and-private"
        )

        # set instance metrics policy
        self.kp.set_instance_metrics_policy(metrics_enable=True)

        # set instance keyCreateImportAccess policy
        self.kp.set_instance_key_create_import_access_policy(
            key_create_import_access_enable=True,
            create_root_key=True,
            create_standard_key=True,
            import_root_key=True,
            import_standard_key=True,
            enforce_token=False,
        )

        # get instance policies
        resp = self.kp.get_instance_policies()

        self.assertEqual(len(resp["resources"]), 4)
        for resource in resp["resources"]:
            if "dualAuthDelete" in resource["policy_type"]:
                self.assertFalse(resource["policy_data"]["enabled"])
            elif "allowedNetwork" in resource["policy_type"]:
                self.assertTrue(resource["policy_data"]["enabled"])
                self.assertEqual(
                    resource["policy_data"]["attributes"]["allowed_network"],
                    "public-and-private",
                )
            elif "metrics" in resource["policy_type"]:
                self.assertTrue(resource["policy_data"]["enabled"])
            elif "keyCreateImportAccess" in resource["policy_type"]:
                self.assertTrue(resource["policy_data"]["enabled"])
                self.assertTrue(
                    resource["policy_data"]["attributes"]["create_root_key"]
                )
                self.assertTrue(
                    resource["policy_data"]["attributes"]["create_standard_key"]
                )
                self.assertTrue(
                    resource["policy_data"]["attributes"]["import_root_key"]
                )
                self.assertTrue(
                    resource["policy_data"]["attributes"]["import_standard_key"]
                )
                self.assertFalse(resource["policy_data"]["attributes"]["enforce_token"])

    def test_set_keyring(self):
        # create a key to be used for test, the key will by default be created in the 'default' key ring
        key = self.kp.create(name="test-key", root=True)
        key_id = key.get("id")
        self.addCleanup(self.kp.delete, key_id)

        # create a key ring to be used for test
        new_key_ring_id = "testKeyRingIdPython"
        self.kp.create_key_ring(new_key_ring_id)

        # transfer key to associate with new key ring
        self.kp.set_key_ring(
            key_id=key_id, key_ring_id="default", new_key_ring_id=new_key_ring_id
        )
        resp = self.kp.get_key(key_id_or_alias=key_id)
        self.assertEqual(resp["keyRingID"], "testKeyRingIdPython")

    @unittest.skipUnless(
        os.environ.get("RUN_PURGE_TESTS") == "true",
        "Skipping purge tests due to lacking permissions",
    )
    def test_purge_key_fail_too_early(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        key_id = key.get("id")
        # delete the key
        self.kp.delete(key_id)
        with self.assertRaises(redstone.client.KeyProtect.KeyProtectError) as cm:
            self.kp.purge_key(key_id=key_id)
        self.assertIn("REQ_TOO_EARLY_ERR", str(cm.exception))

    @unittest.skipUnless(
        os.environ.get("RUN_PURGE_TESTS") == "true",
        "Skipping purge tests due to lacking permissions",
    )
    def test_purge_key_fail_invalid_state(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        key_id = key.get("id")
        self.addCleanup(self.kp.delete, key_id)
        with self.assertRaises(redstone.client.KeyProtect.KeyProtectError) as cm:
            self.kp.purge_key(key_id=key_id)
        self.assertIn("KEY_ACTION_INVALID_STATE_ERR", str(cm.exception))

    def test_sync_resources_fail_too_early(self):
        # create a key to be used for test
        key = self.kp.create(name="test-key", root=True)
        key_id = key.get("id")
        # delete the key
        self.kp.delete(key_id)

        with self.assertRaises(redstone.client.KeyProtect.KeyProtectError) as cm:
            self.kp.sync_associated_resources(key_id=key_id)
        self.assertIn("REQ_TOO_EARLY_ERR", str(cm.exception))

    def test_kmip_happy_path(self):
        # create a key to be used for test
        key = self.kp.create(name="py-sdk-test-key", root=True)
        key_id = key.get("id")

        adapter_created = self.kp.kmip_adapter_create(
            "native_1.0",
            profile_data={
                "crk_id": key_id,
            },
            name="myadapter",
            description="my description",
        )
        adapter_id = adapter_created.get("id")
        self.addCleanup(self.kp.kmip_adapter_delete, adapter_id)

        self.assertEqual(adapter_created["profile"], "native_1.0")

        adapter_getted = self.kp.kmip_adapter_get("myadapter")
        self.assertEqual(adapter_id, adapter_getted["id"])
        self.assertDictEqual(adapter_created, adapter_getted)

        adapters_list = self.kp.kmip_adapter_list(
            filters={"crk_id": key_id},
        )
        self.assertEqual(1, len(adapters_list.get("resources")))
        # print("mdtest-adapterid:", adapter_id)
        cert_created = self.kp.kmip_cert_create(
            adapter_id,
            # adapter_created.get("name"),
            self_signed_cert.generate_selfsigned_cert("www.kms.test.cloud.ibm.com"),
            "mycert",
        )
        cert_getted = self.kp.kmip_cert_get("myadapter", "mycert")
        self.assertDictEqual(cert_created, cert_getted)

        certs_list = self.kp.kmip_cert_list(adapter_id)
        self.assertEqual(1, len(certs_list.get("resources")))

        self.kp.kmip_cert_delete(adapter_id, cert_created.get("id"))


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    unittest.main()



# Copyright 2019 Mathew Odden <mathewrodden@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module holds the service specific client classes, as well
as the BaseClient class that they extend from for shared business function.

If you wish to add or extend functionality to a client or service, this is
where the concrete classes and logic are for those purposes.
"""

import base64
import io
import json
import logging
import re
from typing import Dict, List
import urllib.parse
import zipfile

import requests
import requests.adapters
import requests.auth
from requests.packages.urllib3.util.retry import Retry

from redstone import auth


LOG = logging.getLogger(__name__)


class TokenAuth(requests.auth.AuthBase):
    def __init__(self, credentials):
        self._token_manager = credentials

    def __call__(self, req):
        req.headers["Authorization"] = "Bearer %s" % self._token_manager.get_token()
        return req


class Session(requests.Session):
    def prepare_request(self, request):
        r = super().prepare_request(request)
        for hook in self._pre_send_hooks:
            if hook:
                r = hook(r)
        return r


class BaseClient(object):
    def __init__(
        self,
        region=None,
        service_instance_id=None,
        iam_api_key=None,
        verify=True,
        endpoint_url=None,
        credentials=None,
    ):
        self.session = Session()
        self.session.verify = verify

        self.credentials = credentials
        # respect old path if user builds us directly with API key
        if iam_api_key:
            LOG.warn(
                "'iam_api_key' keyword arg is deprecated. use 'credentials' instead."
            )
            self.credentials = auth.TokenManager(iam_api_key)

        self.session.auth = TokenAuth(self.credentials)

        self.service_instance_id = service_instance_id
        self.region = region

        if endpoint_url:
            if endpoint_url.endswith("/"):
                endpoint_url = endpoint_url[:-1]
            self.endpoint_url = endpoint_url
        else:
            self.endpoint_url = self.endpoint_for_region(region)

        self.session._pre_send_hooks = []

    def set_pre_send_fn(self, fn):
        """
        Sets the provided function as a hook to be called per request,
        just before the requests are sent.

        This allows for modifying any aspects of the requests before they
        are sent, or for recording/logging/tracing purposes.

        The function must handle its own exceptions, and must return a
        `requests.Request` object.

        If `fn` resolves to false-y then it will not be called. This
        can be used to remove any hook currently set, by calling
        `set_pre_send_fn` with `None` type.

        Example::

            def add_header(req):
                req.headers["X-Trace-ID"] = "my-tracing-id"
                return req

            client.set_pre_send_fn(add_header)
            # X-Trace-ID will be set for all requests from this client object
            for i in client.list_instances():
                print(i)

            ...
        """
        if fn is not None:
            self.session._pre_send_hooks = [fn]
        else:
            self.session._pre_send_hooks = []


class IKS(BaseClient):
    names = ["iks"]

    def __init__(self, *args, **kwargs):
        super(IKS, self).__init__(*args, **kwargs)

        # IKS likes to throw back random 503s at times,
        # but retrying generally works fine
        # requests default is Retry(0, read=False); see requests/adapters.py
        retry_conf = Retry(
            total=5, read=False, backoff_factor=1, status_forcelist=[502, 503]
        )
        self.session.mount(
            "https://", requests.adapters.HTTPAdapter(max_retries=retry_conf)
        )
        self.session.mount(
            "http://", requests.adapters.HTTPAdapter(max_retries=retry_conf)
        )

    def endpoint_for_region(self, region):
        return "https://containers.bluemix.net"

    def get_clusters(self) -> List[Dict]:
        """
        List the current IKS clusters in a specific region.

        Returns:
            A list of dict objects representing the cluster metadata.
        """

        """
        GET /v1/clusters HTTP/1.1
        Host: containers.bluemix.net
        Accept: application/json
        Authorization: [PRIVATE DATA HIDDEN]
        Content-Type: application/json
        X-Region: au-syd
        """
        # returns 200 OK on success

        resp = self.session.get(
            "{0}/v1/clusters".format(self.endpoint_url),
            headers={"X-Region": self.region, "Accept": "application/json"},
        )

        if resp.status_code != 200:
            raise Exception(
                "error getting clusters: code=%d body=%r"
                % (resp.status_code, resp.text)
            )

        return resp.json()

    def get_workers(self, cluster):
        """List the workers in an IKS cluster."""

        """
        GET /v1/clusters/<cluster_name_or_id>/workers?showDeleted=false HTTP/1.1
        Host: containers.bluemix.net
        Accept: application/json
        Authorization: [PRIVATE DATA HIDDEN]
        Content-Type: application/json
        X-Region: au-syd
        """
        # returns 200 OK on success

        resp = self.session.get(
            "{0}/v1/clusters/{1}/workers?showDeleted=false".format(
                self.endpoint_url, cluster
            ),
            headers={"X-Region": self.region, "Accept": "application/json"},
        )

        if resp.status_code != 200:
            raise Exception(
                "error getting workers: code=%d body=%r" % (resp.status_code, resp.text)
            )

        return resp.json()

    def update_worker(self, cluster, worker):
        """
        Initiate an update on a worker node.

        The worker node will update to the latest revision that matches the
        master/API server version. i.e. If the master is at 1.16.x, the worker
        will update to the latest 1.16.x series.
        """

        # PUT /v1/clusters/<cluster_name_or_id>/workers/<worker_name_or_id> HTTP/1.1
        # Host: containers.bluemix.net
        # Accept: application/json
        # Authorization: [PRIVATE DATA HIDDEN]
        # Content-Type: application/json
        # X-Region: au-syd

        # {"action":"update"}
        # OR
        # {"action":"update", "force": true}
        # returns 204 No Content on success

        resp = self.session.put(
            "{0}/v1/clusters/{1}/workers/{2}".format(
                self.endpoint_url, cluster, worker
            ),
            headers={"X-Region": self.region, "Accept": "application/json"},
            json={"action": "update"},
        )

        if resp.status_code != 204:
            raise Exception(
                "error updating workers: code=%d body=%r"
                % (resp.status_code, resp.text)
            )

    def update_master(self, cluster, version):
        """Initate an update on the master nodes of a cluster."""

        resp = self.session.put(
            "{0}/v1/clusters/{1}".format(self.endpoint_url, cluster),
            headers={"X-Region": self.region, "Accept": "application/json"},
            json={"action": "update", "version": version},
        )

        if resp.status_code != 204:
            raise Exception(
                "error updating master: code=%d body=%r" % (resp.status_code, resp.text)
            )

    def get_kube_versions(self):
        resp = self.session.get(
            "{0}/v1/kube-versions".format(self.endpoint_url),
            headers={"X-Region": self.region, "Accept": "application/json"},
        )

        if resp.status_code != 200:
            raise Exception(
                "error getting kube-versions: code=%d body=%r"
                % (resp.status_code, resp.text)
            )

        return resp.json()

    def get_cluster_config(self, cluster):
        """
        Retrieve a KubeConfig that can be used with kubectl to
        interact with the given IKS cluster.

        Returns:
            base64 encoded file data; can be decode and written to file,
            or used with the python kubernetes client to interact in the same process
        """
        output_format = "yaml"

        # NOTE(mrodden): no idea why this one API needs the refresh token,
        # but it certainly gives an HTTP 400 when you don't have it
        # this is a pretty hacky way of getting one, but we shouldn't even
        # need it either so... idk. /shrug

        # do a get token before, to make sure we get a valid refresh token
        _ = self.session.auth._token_manager.get_token()
        refresh_token = self.session.auth._token_manager._token_info.get(
            "refresh_token"
        )

        # yaml output was added after our original code to deal with the zipfile
        # leaving both in because maybe its useful for someone to use the zip path still
        params = {}
        if output_format == "yaml":
            params["format"] = "yaml"

        path = "{0}/v1/clusters/{1}/config"
        resp = self.session.get(
            path.format(self.endpoint_url, cluster),
            params=params,
            headers={"X-Region": self.region, "X-Auth-Refresh-Token": refresh_token},
        )

        resp.raise_for_status()

        if output_format == "yaml":
            config_yaml = resp.content
        else:
            zip_data = io.BytesIO()
            zip_data.write(resp.content)
            zip_file = zipfile.ZipFile(zip_data)

            files = zip_file.namelist()
            for file_name in files:
                if file_name.endswith("yml"):
                    kube_yaml = zip_file.read(file_name)
                elif file_name.endswith("pem"):
                    ca_data = zip_file.read(file_name)

            ca_b64 = base64.b64encode(ca_data)

            config_yaml = re.sub(
                b"certificate-authority: .+",
                b"certificate-authority-data: " + ca_b64,
                kube_yaml,
            )

        return base64.b64encode(config_yaml)


class ResourceController(BaseClient):
    """
    Client class for interacting with the Resource Controller service,
    which is used for managing service instances within the cloud account.

    API Docs:
      - https://console.bluemix.net/apidocs/resource-manager
      - https://console.bluemix.net/apidocs/resource-controller
    """

    names = ["rc"]
    KEYPROTECT_PLAN_ID = (
        "eedd3585-90c6-4c8f-be3d-062069e99fc3"  # keyprotect tiered-pricing ID
    )

    def __init__(self, *args, **kwargs):
        super(ResourceController, self).__init__(*args, **kwargs)

        # add retries for RC, since it has issues communicating with global catalog
        # and will throw a 504 in this case
        retry_conf = Retry(
            total=5, read=False, backoff_factor=1, status_forcelist=[502, 503, 504]
        )
        self.session.mount(
            "https://", requests.adapters.HTTPAdapter(max_retries=retry_conf)
        )
        self.session.mount(
            "http://", requests.adapters.HTTPAdapter(max_retries=retry_conf)
        )

    def endpoint_for_region(self, region):
        return "https://resource-controller.cloud.ibm.com"

    def get_default_resource_group(self):
        default_rg = next(
            filter(
                lambda x: x.get("name") in ["Default", "default"],
                self.resource_groups().get("resources", []),
            ),
            None,
        )

        if not default_rg:
            raise Exception("No default resource group found!")

        return default_rg

    def resource_groups(self):
        # resource-manager used to be independent when it was bluemix,
        # but any recent endpoint will be the same for controller and manager
        if "bluemix" in self.endpoint_url:
            netloc = self.endpoint_url.replace("controller", "manager")
        else:
            netloc = self.endpoint_url

        # apparently it doesn't complain if we drop query params,
        # didn't want to have to look up the account ID anyway, so +2
        resp = self.session.get(
            "{0}/v2/resource_groups".format(netloc),
        )

        if resp.status_code != 200:
            raise Exception(
                "Failed to get resource groups: url=%r code=%d body=%r"
                % (resp.request.url, resp.status_code, resp.text)
            )

        return resp.json()

    def create_instance(self, name, plan_id, region=None, resource_group=None):
        """
        Create/provision a service instance.

        Returns:
            tuple of (service_GUID, service_CRN) if successful

        Raises:
            Exception if there is an error
        """

        if resource_group is None:
            resource_group_id = self.get_default_resource_group().get("id")
        else:
            resource_group_id = resource_group

        if not region:
            region = self.region

        return self._create_instance_v2(name, region, resource_group_id, plan_id)

    def _create_instance_v2(self, name, region, resource_group_id, resource_plan_id):
        body = {
            "name": name,
            "resource_plan_id": resource_plan_id,
            "resource_group": resource_group_id,
            "target": region,
        }

        resp = self.session.post(
            "{0}/v2/resource_instances".format(self.endpoint_url), json=body
        )

        if resp.status_code != 201:
            raise Exception(
                "Create instance failed: code=%d body=%s"
                % (resp.status_code, resp.text)
            )

        return resp.json().get("guid"), resp.json().get("id")

    def _create_instance_v1(self, name, region, resource_group_id, resource_plan_id):
        # seems like the target_crn is the region selector,
        # and its just the price plan ID with the region stuck at the end
        target_crn = (
            "crn:v1:bluemix:public:globalcatalog::::deployment:{0}%3A{1}".format(
                resource_plan_id, region
            )
        )

        body = {
            "name": name,
            "resource_plan_id": resource_plan_id,
            "resource_group_id": resource_group_id,
            "target_crn": target_crn,
        }

        resp = self.session.post(
            "{0}/v1/resource_instances".format(self.endpoint_url), json=body
        )

        if resp.status_code != 201:
            raise Exception(
                "Create instance failed: code=%d body=%s"
                % (resp.status_code, resp.text)
            )

        return resp.json().get("guid"), resp.json().get("id")

    def delete_instance(self, instance_crn):
        """Delete/deprovision a service instance identified by the given CRN or UUID."""

        safe_crn = urllib.parse.quote(instance_crn, "")
        resp = self.session.delete(
            "{0}/v2/resource_instances/{1}".format(self.endpoint_url, safe_crn)
        )

        if resp.status_code != 204:
            raise Exception(
                "Delete instance failed: code=%d body=%s"
                % (resp.status_code, resp.text)
            )

    def list_instances(self):
        """
        Retrieve a list of all service and resource instances in the current account.

        Note this will return an iterator that will handle the underlying pagination of
        large sets of instances returned.

        Returns:
            a generator type that iterates over the collection of instances returned from
            the API request
        """
        resp = self.session.get("{0}/v2/resource_instances".format(self.endpoint_url))
        resp.raise_for_status()

        while True:
            for res in resp.json()["resources"]:
                yield res

            next_url = resp.json().get("next_url")
            if not next_url:
                break

            resp = self.session.get("{0}{1}".format(self.endpoint_url, next_url))
            resp.raise_for_status()

    def get_instance(self, instance_id):
        resp = self.session.get(
            "{0}/v2/resource_instances/{1}".format(self.endpoint_url, instance_id)
        )
        if resp.status_code == 404:
            return None

        resp.raise_for_status()
        return resp.json()


class KeyProtect(BaseClient):
    """
    API Docs: https://cloud.ibm.com/apidocs/key-protect
    """

    class KeyProtectError(Exception):
        @staticmethod
        def wrap(http_error):
            try:
                message = http_error.response.json()["resources"][0]["errorMsg"]
            except (KeyError, json.decoder.JSONDecodeError, ValueError):
                message = http_error.response.text
            err = KeyProtect.KeyProtectError(message)
            err.http_error = http_error
            err.__suppress_context__ = True
            return err

    names = ["kms"]

    def __init__(self, *args, **kwargs):
        super(KeyProtect, self).__init__(*args, **kwargs)

        if not self.service_instance_id:
            raise ValueError(
                "KeyProtect service requires 'service_instance_id' to be set!"
            )

        self.session.headers["Bluemix-Instance"] = self.service_instance_id

    def endpoint_for_region(self, region):
        return "https://{0}.kms.cloud.ibm.com".format(region)

    def _validate_resp(self, resp):
        def log_resp(resp):
            resp_str = io.StringIO()
            print("%s %s" % (resp.status_code, resp.reason), file=resp_str)

            for k, v in resp.headers.items():
                if k.lower() == "authorization":
                    v = "REDACTED"
                print("%s: %s" % (k, v), file=resp_str)

            print(resp.content.decode(), end="", file=resp_str)
            return resp_str.getvalue()

        try:
            LOG.debug(log_resp(resp))
            resp.raise_for_status()
        except requests.HTTPError as http_err:
            http_err.raw_response = log_resp(resp)
            raise KeyProtect.KeyProtectError.wrap(http_err)

    def list_keys(self):
        resp = self.session.get("%s/api/v2/keys" % self.endpoint_url)

        self._validate_resp(resp)

        return resp.json().get("resources", [])

    def get_key(self, key_id_or_alias):
        resp = self.session.get(
            "%s/api/v2/keys/%s" % (self.endpoint_url, key_id_or_alias)
        )

        self._validate_resp(resp)

        return resp.json().get("resources")[0]

    def create_key(
        self, name, payload=None, raw_payload=None, root=False, alias_list=None
    ):
        data = {
            "metadata": {
                "collectionType": "application/vnd.ibm.kms.key+json",
                "collectionTotal": 1,
            },
            "resources": [
                {
                    "type": "application/vnd.ibm.kms.key+json",
                    "extractable": not root,
                    "name": name,
                }
            ],
        }

        # use raw_payload if given, else assume payload needs some base64 love
        if raw_payload is not None:
            data["resources"][0]["payload"] = raw_payload
        elif payload is not None:
            data["resources"][0]["payload"] = base64.b64encode(payload).decode("utf-8")

        # creates a new key with alias name. A key can have a maximum of 5 alias names
        if alias_list:
            if len(alias_list) > 5:
                raise ValueError("A key can not have more than 5 alias names")
            else:
                data["resources"][0]["aliases"] = alias_list

        resp = self.session.post("%s/api/v2/keys" % self.endpoint_url, json=data)
        self._validate_resp(resp)
        return resp.json().get("resources")[0]

    def delete_key(self, key_id):
        resp = self.session.delete("%s/api/v2/keys/%s" % (self.endpoint_url, key_id))
        self._validate_resp(resp)

    def _action(self, key_id, action, jsonable):
        resp = self.session.post(
            "%s/api/v2/keys/%s/actions/%s" % (self.endpoint_url, key_id, action),
            json=jsonable,
        )
        self._validate_resp(resp)
        if resp.status_code == 204:
            return None
        return resp.json()

    def wrap(self, key_id, plaintext, aad=None):
        if plaintext:
            data = {"plaintext": base64.b64encode(plaintext).decode("utf-8")}
        else:
            data = {}

        if aad:
            data["aad"] = aad

        return self._action(key_id, "wrap", data)

    def unwrap(self, key_id, ciphertext, aad=None):
        # json body needs to be a UTF-8 string
        if isinstance(ciphertext, bytes):
            ciphertext = ciphertext.decode("utf-8")

        data = {"ciphertext": ciphertext}

        if aad:
            data["aad"] = aad

        resp = self._action(key_id, "unwrap", data)
        return base64.b64decode(resp["plaintext"].encode("utf-8"))

    def rewrap(self, key_id, ciphertext, aad=None):
        # json body needs to be a UTF-8 string
        if isinstance(ciphertext, bytes):
            ciphertext = ciphertext.decode("utf-8")

        data = {"ciphertext": ciphertext}

        if aad:
            data["aad"] = aad

        resp = self._action(key_id, "rewrap", data)
        return resp

    def rotate_key(self, key_id, payload=None):
        data = None
        if payload:
            data = {"payload": base64.b64encode(payload).decode("utf-8")}

        return self._action(key_id, "rotate", data)

    def restore_key(self, key_id: str):
        """
        Restore a key.

        The RestoreKey method reverts a key's status from `Destroyed` to `Active`.
        This method cannot be used to restore a key that has been purged.

        API Docs: https://cloud.ibm.com/apidocs/key-protect#restorekey
        """
        resp = self.session.post(
            "%s/api/v2/keys/%s/restore" % (self.endpoint_url, key_id)
        )
        self._validate_resp(resp)

    def disable_key(self, key_id: str):
        """
        Disable a key.

        The key will not be deleted, but it will not be active and
        key operations cannot be performed on a disabled key.

        API Docs: https://cloud.ibm.com/apidocs/key-protect#disablekey
        """
        resp = self.session.post(
            "%s/api/v2/keys/%s/actions/disable" % (self.endpoint_url, key_id)
        )
        self._validate_resp(resp)

    def enable_key(self, key_id: str):
        """
        Enable a key.

        Only disabled keys can be enabled. After calling this action, the
        key becomes active and key operations can be performed on it.

        Note: This does not recover Deleted keys.

        API Docs: https://cloud.ibm.com/apidocs/key-protect#enablekey
        """
        resp = self.session.post(
            "%s/api/v2/keys/%s/actions/enable" % (self.endpoint_url, key_id)
        )
        self._validate_resp(resp)

    def sync_associated_resources(self, key_id: str):
        """
        Executes the sync request which verifies and updates the resources associated with the key.

        API Docs: https://cloud.ibm.com/apidocs/key-protect#syncassociatedresources
        """

        resp = self.session.post(
            "%s/api/v2/keys/%s/actions/sync" % (self.endpoint_url, key_id),
        )
        self._validate_resp(resp)

    def create_key_ring(self, key_ring_id: str):
        """
        Create a key ring in the instance with the specified name.

        API Docs: https://cloud.ibm.com/apidocs/key-protect#createkeyring
        """

        resp = self.session.post(
            "%s/api/v2/key_rings/%s" % (self.endpoint_url, key_ring_id),
        )
        self._validate_resp(resp)

    def get_key_rings(self):
        """
        Get all key rings associated with specified instance

        API Docs: https://cloud.ibm.com/apidocs/key-protect#listkeyrings
        """
        resp = self.session.get(
            "%s/api/v2/key_rings" % self.endpoint_url,
        )
        self._validate_resp(resp)
        return resp.json()

    def delete_key_ring(self, key_ring_id: str):
        """
        Deletes a key ring from the associated instance

        API Docs: https://cloud.ibm.com/apidocs/key-protect#deletekeyring
        """
        resp = self.session.delete(
            "%s/api/v2/key_rings/%s" % (self.endpoint_url, key_ring_id),
        )
        self._validate_resp(resp)

    def create_import_token(
        self, expiration: int = None, max_allowed_retrievals: int = None
    ):
        """
        Create an import token that can be used to import encrypted material as root keys.

        expiration: The time in seconds from the creation of a import token that determines how long it remains valid.
            The minimum value is 300 seconds (5 minutes), and the maximum value is 86400 (24 hours). The default value
            is 600 (10 minutes).
        max_allowed_retrievals: The number of times that an import token can be retrieved within its expiration time
            before it is no longer accessible. The default value is 1. The maximum value is 500.

        API Docs: https://cloud.ibm.com/apidocs/key-protect#postimporttoken
        """

        data = {}
        if expiration:
            data["expiration"] = expiration
        if max_allowed_retrievals:
            data["maxAllowedRetrievals"] = max_allowed_retrievals
        resp = self.session.post(
            "%s/api/v2/import_token" % self.endpoint_url, json=data
        )

        self._validate_resp(resp)
        return resp.json()

    def get_registrations(self, key_id: str = None, crn: str = None):
        """
        Retrieve a list of registrations

        If `key_id` is None (the default) all registrations for the instance are
        returned, otherwise only the registrations associated with a specified root
        key are returned.

        `crn` should be a str type that will be passed as the `urlEncodedResourceCRNQuery` parameter to the HTTP API.
        It is used to filter registration on a specific cloud resource. More information can be found in the API docs below.

        API Docs:
          - https://cloud.ibm.com/apidocs/key-protect#getregistrations
          - https://cloud.ibm.com/apidocs/key-protect#getregistrationsallkeys
        """
        params = {}
        if crn is not None:
            params["urlEncodedResourceCRNQuery"] = crn

        if key_id is not None:
            url = "%s/api/v2/keys/%s/registrations" % (self.endpoint_url, key_id)
        else:
            url = "%s/api/v2/keys/registrations" % self.endpoint_url

        resp = self.session.get(url, params=params)

        self._validate_resp(resp)
        return resp.json()

    def get_import_token(self):
        """
        Retrieves an import token associated with the current service instance. Token must be previously created by
        a create import token call.

        API Docs: https://cloud.ibm.com/apidocs/key-protect#getimporttoken
        """

        resp = self.session.get("%s/api/v2/import_token" % self.endpoint_url)
        self._validate_resp(resp)
        return resp.json()

    def create_key_alias(self, key_id: str, alias: str):
        """
        Creates an alias name for a key.
        An alias is a user defined string that can be used in place of a normal UUID Key ID

        API Docs: https://cloud.ibm.com/apidocs/key-protect#createkeyalias
        """

        resp = self.session.post(
            "%s/api/v2/keys/%s/aliases/%s" % (self.endpoint_url, key_id, alias)
        )
        self._validate_resp(resp)
        return resp.json()

    def delete_key_alias(self, key_id: str, alias: str):
        """
        Deletes an alias name associated with a key

        API Docs: https://cloud.ibm.com/apidocs/key-protect#deletekeyalias
        """

        resp = self.session.delete(
            "%s/api/v2/keys/%s/aliases/%s" % (self.endpoint_url, key_id, alias)
        )
        self._validate_resp(resp)

    def _set_policy(self, resources_list, key_id=None):
        collection_total = len(resources_list)
        data = {
            "metadata": {
                "collectionType": "application/vnd.ibm.kms.policy+json",
                "collectionTotal": collection_total,
            },
            "resources": resources_list,
        }

        if key_id:
            resp = self.session.put(
                "%s/api/v2/keys/%s/policies" % (self.endpoint_url, key_id), json=data
            )
        else:
            resp = self.session.put(
                "%s/api/v2/instance/policies" % self.endpoint_url, json=data
            )

        self._validate_resp(resp)
        if resp.status_code == 204:
            return
        else:
            return resp.json()

    def set_key_rotation_policy(self, key_id: str, rotation_interval: int):
        """
        Updates the rotation policy associated with a key by specifying key ID and rotation interval

        API Docs: https://cloud.ibm.com/apidocs/key-protect#putpolicy
        """
        resources_list = [
            {
                "type": "application/vnd.ibm.kms.policy+json",
                "rotation": {
                    "interval_month": int(rotation_interval),
                },
            }
        ]
        return self._set_policy(resources_list, key_id)

    def set_key_dual_auth_policy(self, key_id: str, dual_auth_enable: bool):
        """
        Updates the dual auth delete policy by passing the key ID and enable detail

        API Docs: https://cloud.ibm.com/apidocs/key-protect#putpolicy
        """
        resources_list = [
            {
                "type": "application/vnd.ibm.kms.policy+json",
                "dualAuthDelete": {"enabled": dual_auth_enable},
            }
        ]
        return self._set_policy(resources_list, key_id)

    def get_key_policies(self, key_id: str):
        """
        Retrieves a list of policies that are associated with a specified key

        API Docs: https://cloud.ibm.com/apidocs/key-protect#getpolicy
        """
        resp = self.session.get(
            "%s/api/v2/keys/%s/policies" % (self.endpoint_url, key_id)
        )
        self._validate_resp(resp)
        return resp.json()

    def initiate_dual_auth_delete(self, key_id: str):
        """
        Authorize deletion for a key with a dual authorization policy

        API Docs: https://cloud.ibm.com/apidocs/key-protect#setkeyfordeletion
        """
        resp = self.session.post(
            "%s/api/v2/keys/%s/actions/setKeyForDeletion" % (self.endpoint_url, key_id)
        )
        self._validate_resp(resp)

    def cancel_dual_auth_delete(self, key_id: str):
        """
        Remove an authorization for a key with a dual authorization policy

        API Docs: https://cloud.ibm.com/apidocs/key-protect#unsetkeyfordeletion
        """
        resp = self.session.post(
            "%s/api/v2/keys/%s/actions/unsetKeyForDeletion"
            % (self.endpoint_url, key_id)
        )
        self._validate_resp(resp)

    def set_instance_dual_auth_policy(self, dual_auth_enable: bool):
        """
        Updates the dual auth delete policy for the instance by passing the enable detail

        API Docs: https://cloud.ibm.com/apidocs/key-protect#putinstancepolicy
        """
        resources_list = [
            {
                "policy_type": "dualAuthDelete",
                "policy_data": {"enabled": dual_auth_enable},
            }
        ]
        return self._set_policy(resources_list)

    def set_instance_allowed_network_policy(
        self, allowed_network_enable: bool, network_type: str
    ):
        """
        Updates the allowed network policy for the instance

        `network_type` is a str type, and must be one of "public-and-private" or "private-only".
        The default is "public-and-private", but can be set to "private-only" to disable access to the instance from Internet client addresses.
        API Docs: https://cloud.ibm.com/apidocs/key-protect#putinstancepolicy
        """
        resources_list = [
            {
                "policy_type": "allowedNetwork",
                "policy_data": {
                    "enabled": allowed_network_enable,
                    "attributes": {"allowed_network": network_type},
                },
            }
        ]
        return self._set_policy(resources_list)

    def set_instance_allowed_ip_policy(
        self, allowed_ip_enable: bool, allowed_ips: List[str]
    ):
        """
        Updates the allowed IP policy for the instance

        API Docs: https://cloud.ibm.com/apidocs/key-protect#putinstancepolicy
        """
        resources_list = [
            {
                "policy_type": "allowedIP",
                "policy_data": {
                    "enabled": allowed_ip_enable,
                    "attributes": {"allowed_ip": allowed_ips},
                },
            }
        ]
        return self._set_policy(resources_list)

    def set_instance_key_create_import_access_policy(
        self,
        key_create_import_access_enable: bool,
        create_root_key: bool,
        create_standard_key: bool,
        import_root_key: bool,
        import_standard_key: bool,
        enforce_token: bool,
    ):
        """
        Updates the key create import access policy details associated with an instance.

        'key_create_import_access_enable' is boolean type, the instance policy is enabled if it's set to true

        API Docs: https://cloud.ibm.com/apidocs/key-protect#putinstancepolicy
        """
        resources_list = [
            {
                "policy_type": "keyCreateImportAccess",
                "policy_data": {
                    "enabled": key_create_import_access_enable,
                    "attributes": {
                        "create_root_key": create_root_key,
                        "create_standard_key": create_standard_key,
                        "import_root_key": import_root_key,
                        "import_standard_key": import_standard_key,
                        "enforce_token": enforce_token,
                    },
                },
            }
        ]
        return self._set_policy(resources_list)

    def set_instance_metrics_policy(self, metrics_enable: bool):
        """
        Updates the metrics policy details associated with an instance

        API Docs: https://cloud.ibm.com/apidocs/key-protect#putinstancepolicy
        """
        resources_list = [
            {
                "policy_type": "metrics",
                "policy_data": {"enabled": metrics_enable},
            }
        ]
        return self._set_policy(resources_list)

    def get_instance_policies(self):
        """
        Retrieves a list of policies that are associated with a specified service instance

        https://cloud.ibm.com/apidocs/key-protect#getinstancepolicy
        """
        resp = self.session.get("%s/api/v2/instance/policies" % self.endpoint_url)
        self._validate_resp(resp)
        return resp.json()

    def set_key_ring(self, key_id: str, key_ring_id: str, new_key_ring_id: str):
        """
        Transfers a key associated with one key ring to another key ring

        https://cloud.ibm.com/apidocs/key-protect#patchkey
        """
        resp = self.session.patch(
            "%s/api/v2/keys/%s" % (self.endpoint_url, key_id),
            json={"keyRingID": new_key_ring_id},
            headers={"X-Kms-Key-Ring": key_ring_id},
        )
        self._validate_resp(resp)

    def purge_key(self, key_id: str):
        """
        Purge key method shreds all the metadata and registrations associated with a key that has been
        deleted. The purge operation is allowed to be performed on a key from 4 hours after its deletion

        https://cloud.ibm.com/apidocs/key-protect#purgekey
        """
        resp = self.session.delete(
            "%s/api/v2/keys/%s/purge" % (self.endpoint_url, key_id)
        )
        self._validate_resp(resp)

    def kmip_adapter_create(
        self,
        profile: str,
        profile_data: dict,
        name: str = None,
        description: str = None,
    ):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#create-kmip-adapter
        """
        adapter_to_create = {"profile": profile, "profile_data": profile_data}
        if name is not None:
            adapter_to_create["name"] = name
        if description is not None:
            adapter_to_create["description"] = description
        resp = self.session.post(
            f"{self.endpoint_url}/api/v2/kmip_adapters",
            json={
                "metadata": {
                    "collectionTotal": 1,
                    "collectionType": "application/vnd.ibm.kms.kmip_adapter+json",
                },
                "resources": [adapter_to_create],
            },
        )
        self._validate_resp(resp)
        return resp.json().get("resources")[0]

    def kmip_adapter_delete(self, adapter_name_or_id: str):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#delete-kmip-adapter
        """
        resp = self.session.delete(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}",
        )
        self._validate_resp(resp)

    def kmip_adapter_get(self, adapter_name_or_id: str):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#get-kmip-adapter
        """
        resp = self.session.get(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}",
        )
        self._validate_resp(resp)
        return resp.json().get("resources")[0]

    def kmip_adapter_list(
        self,
        limit: int = 200,
        offset: int = 0,
        show_total: bool = False,
        filters: dict = None,
    ):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#get-kmip-adapters
        """
        params = {"limit": limit, "offset": offset, "totalCount": show_total}
        if filters is not None:
            if "crk_id" in filters:
                params["crk_id"] = filters["crk_id"]
        resp = self.session.get(
            f"{self.endpoint_url}/api/v2/kmip_adapters", params=params
        )
        self._validate_resp(resp)
        return resp.json()

    def kmip_cert_create(
        self, adapter_name_or_id: str, cert_payload: str, cert_name: str = None
    ):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#create-kmip-client-certificate
        """
        cert_to_create = {"certificate": cert_payload}
        if cert_name is not None:
            cert_to_create["name"] = cert_name

        resp = self.session.post(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}/certificates",
            json={
                "metadata": {
                    "collectionTotal": 1,
                    "collectionType": "application/vnd.ibm.kms.kmip_client_certificate+json",
                },
                "resources": [cert_to_create],
            },
        )
        self._validate_resp(resp)
        return resp.json().get("resources")[0]

    def kmip_cert_delete(self, adapter_name_or_id: str, cert_name_or_id: str):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#delete-kmip-client-certificate
        """
        resp = self.session.delete(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}/certificates/{cert_name_or_id}",
        )
        self._validate_resp(resp)

    def kmip_cert_get(self, adapter_name_or_id: str, cert_name_or_id: str):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#get-kmip-client-certificate
        """
        resp = self.session.get(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}/certificates/{cert_name_or_id}",
        )
        self._validate_resp(resp)
        return resp.json().get("resources")[0]

    def kmip_cert_list(
        self,
        adapter_name_or_id: str,
        limit: int = 200,
        offset: int = 0,
        show_total: bool = False,
    ):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#get-kmip-client-certificates
        """
        params = {"limit": limit, "offset": offset, "totalCount": show_total}
        resp = self.session.get(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}/certificates",
            params=params,
        )
        self._validate_resp(resp)
        return resp.json()

    def kmip_object_delete(self, adapter_name_or_id: str, obj_id: str):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#delete-kmip-object
        """
        resp = self.session.delete(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}/kmip_objects/{obj_id}",
        )
        self._validate_resp(resp)

    def kmip_object_get(self, adapter_name_or_id: str, obj_id: str):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#get-kmip-object
        """
        resp = self.session.get(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}/kmip_objects/{obj_id}",
        )
        self._validate_resp(resp)
        return resp.json().get("resources")[0]

    def kmip_object_list(
        self,
        adapter_name_or_id: str,
        limit: int = 200,
        offset: int = 0,
        show_total: bool = False,
        state_filter: list = [1, 2, 3, 4],
    ):
        """
        API Docs: https://cloud.ibm.com/apidocs/key-protect#get-kmip-objects
        """
        params = {
            "limit": limit,
            "offset": offset,
            "totalCount": show_total,
            "state": state_filter,
        }
        resp = self.session.get(
            f"{self.endpoint_url}/api/v2/kmip_adapters/{adapter_name_or_id}/kmip_objects",
            params=params,
        )
        self._validate_resp(resp)
        return resp.json()

    # deprecated methods
    keys = list_keys
    get = get_key
    create = create_key
    delete = delete_key
    rotate = rotate_key


class CISAuth(requests.auth.AuthBase):
    def __init__(self, credentials):
        self._token_manager = credentials

    def __call__(self, req):
        req.headers["x-auth-user-token"] = "Bearer %s" % self._token_manager.get_token()
        return req


class CIS(BaseClient):
    names = ["cis"]

    def __init__(self, *args, **kwargs):
        super(CIS, self).__init__(*args, **kwargs)
        self.session.auth = CISAuth(self.credentials)
        self.safe_crn = urllib.parse.quote(self.service_instance_id, safe="")

    def endpoint_for_region(self, region):
        return "https://api.cis.cloud.ibm.com"

    def pools(self):
        path = "{0}/v1/{1}/load_balancers/pools"
        resp = self.session.get(path.format(self.endpoint_url, self.safe_crn))
        resp.raise_for_status()
        return resp.json().get("result")

    def get_pool(self, pool_id):
        path = "{0}/v1/{1}/load_balancers/pools/{2}"
        resp = self.session.get(path.format(self.endpoint_url, self.safe_crn, pool_id))
        resp.raise_for_status()
        return resp.json().get("result")

    def update_pool(self, pool):
        pool_id = pool.get("id")

        keys_to_remove = ["created_on", "modified_on", "healthy", "id"]

        for key in keys_to_remove:
            try:
                del pool[key]
            except KeyError:
                pass

        path = "{0}/v1/{1}/load_balancers/pools/{2}"
        resp = self.session.put(
            path.format(self.endpoint_url, self.safe_crn, pool_id), json=pool
        )
        resp.raise_for_status()
        return resp.json().get("result")



# Copyright 2020 Mathew Odden <mathewrodden@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import base64
import json
import logging
import os
import sys

import redstone


def lookup_instance(instance_id):
    rc = redstone.service("ResourceController")
    instance = rc.get_instance(instance_id)
    return instance


def pp_json(thing):
    print(json.dumps(thing, indent=4, sort_keys=True))


def _main():
    p = argparse.ArgumentParser()
    p.add_argument("--instance-id", "-i")

    subp = p.add_subparsers(dest="action")

    _ = subp.add_parser("list-instances")

    _ = subp.add_parser("list")

    cp = subp.add_parser("create")
    cp.add_argument("name")
    cp.add_argument("--data")
    cp.add_argument("--exportable", action="store_true")

    gp = subp.add_parser("get")
    gp.add_argument("key_id")

    dp = subp.add_parser("delete")
    dp.add_argument("key_id", nargs="+")

    wp = subp.add_parser("wrap")
    wp.add_argument("key_id")
    wp.add_argument("--data")
    wp.add_argument("--aad")

    uwp = subp.add_parser("unwrap")
    uwp.add_argument("key_id")
    uwp.add_argument("data")
    uwp.add_argument("--aad")
    uwp.add_argument("--override-handle")

    rotate_p = subp.add_parser("rotate")
    rotate_p.add_argument("key_id")
    rotate_p.add_argument("--data")

    args = p.parse_args()

    if not args.action:
        p.print_usage()

    if args.action == "list-instances":
        rc = redstone.service("ResourceController")

        instances = rc.list_instances()
        kms_instances = list(
            filter(
                lambda x: x["type"] == "service_instance"
                and x.get("sub_type") == "kms",
                instances,
            )
        )

        render_list(
            kms_instances,
            fields=["guid", "name", "region_id"],
            titles=["ID", "NAME", "REGION"],
        )
        return

    instance_id = args.instance_id or os.environ.get("KP_INSTANCE_ID")
    if not instance_id:
        print("No instance ID given, set KP_INSTANCE_ID or pass --instance-id")
        return 1

    instance_data = lookup_instance(instance_id)
    if not instance_data:
        print("No instace found for ID: %s" % instance_id)
        return 1

    kp = redstone.service(
        "KeyProtect", service_instance_id=instance_id, region=instance_data["region_id"]
    )

    if args.action == "list":
        fields = ["id", "name", "extractable"]
        render_list(kp.keys(), fields=fields, titles=map(lambda x: x.upper(), fields))
    elif args.action == "create":
        key_data = None
        if args.data:
            key_data = args.data
        key = kp.create(args.name, raw_payload=key_data, root=not args.exportable)
        pp_json(key)
    elif args.action == "get":
        key = kp.get(args.key_id)
        pp_json(key)
    elif args.action == "delete":
        for key_id in args.key_id:
            resp = kp.delete(key_id)
            print("Deleted key %s" % key_id)
    elif args.action == "wrap":
        plaintext = None
        if args.data:
            plaintext = args.data.encode()
        resp = kp.wrap(args.key_id, plaintext, args.aad)
        pp_json(resp)
    elif args.action == "unwrap":
        if args.override_handle:
            unpacked = json.loads(base64.b64decode(args.data.encode()).decode())
            unpacked["handle"] = args.override_handle
            data = base64.b64encode(json.dumps(unpacked).encode()).decode()
        else:
            data = args.data
        res = kp.unwrap(args.key_id, data, args.aad)
        print(base64.b64encode(res).decode())
    elif args.action == "rotate":
        new_key_data = None
        if args.data:
            new_key_data = args.data.encode()
        res = kp.rotate(args.key_id, new_key_data)
        print(res)


def render_list(data, fields, titles):

    lengths = {}

    for datum in data:
        for field in fields:
            lengths[field] = max(len(str(datum.get(field, 0))), lengths.get(field, 0))

    fmt_str = " ".join(map(lambda x: "%%(%s)-%ds" % (x, lengths.get(x, 0) + 3), fields))

    print(fmt_str % dict(zip(fields, titles)))
    for datum in data:
        print(fmt_str % datum)


def main():
    logging.basicConfig(level=logging.INFO)
    sys.exit(_main())


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import patch

from redstone import auth


class TokenManagerTestCase(unittest.TestCase):
    def test_gettoken_expired_norefreshtoken(self):
        """
        When get_token is called, the current token is expired,
        and we are not using refresh tokens,
        then we should not call _refresh_token.
        """

        p1 = patch.object(auth.TokenManager, "is_token_expired", return_value=True)
        p2 = patch.object(auth.TokenManager, "_refresh_token")
        p3 = patch.object(auth.TokenManager, "_request_token")
        p4 = patch.object(
            auth.TokenManager, "is_refresh_token_expired", return_value=False
        )
        ite = p1.start()
        ref_tok = p2.start()
        req_tok = p3.start()
        _ = p4.start()
        self.addCleanup(p1.stop)
        self.addCleanup(p2.stop)
        self.addCleanup(p3.stop)
        self.addCleanup(p4.stop)

        tm = auth.TokenManager(api_key="TEST_KEY", use_refresh_token=False)
        tm._token_info["access_token"] = "MOCK_TOKEN"
        tm.get_token()

        ite.assert_called()
        ref_tok.assert_not_called()
        req_tok.assert_called()



# Copyright 2020 Mathew Odden <mathewrodden@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import json
import logging
import os
from typing import List, Optional, Tuple

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

import redstone
import redstone.crn


LOG = logging.getLogger(__name__)


class MessageHeader(object):

    version = 1

    def __init__(self, data_keys, aad, algorithm):
        self.data_keys = data_keys
        self.aad = aad
        self.algorithm = algorithm

    def _pack(self):
        message_header_data = {
            "data_keys": self.data_keys,
            "aad": self.aad,
            "algorithm": self.algorithm,
        }
        message_header = base64.b64encode(
            json.dumps(message_header_data).encode("utf-8")
        )
        message_header = (
            _message_version_to_bytes(self.version)
            + len(message_header).to_bytes(8, byteorder="big")
            + message_header
        )
        return message_header

    @staticmethod
    def from_message_with_body(message: bytes) -> "Tuple[MessageHeader, bytes]":
        version = get_message_version(message)
        if version != MessageHeader.version:
            raise Exception(
                "Invalid message version. Expecting %d, found %d"
                % (MessageHeader.version, version)
            )

        header_len = int.from_bytes(message[1:9], byteorder="big")
        header_bytes = message[9 : header_len + 9]
        header_dict = json.loads(base64.b64decode(header_bytes).decode("utf-8"))
        return MessageHeader(**header_dict), message[header_len + 9 :]

    def __repr__(self):
        prop_strs = [
            "%s=%r" % (key, getattr(self, key))
            for key in ["version", "data_keys", "aad", "algorithm"]
        ]
        return "<%s %s>" % (type(self).__name__, ", ".join(prop_strs))


def encrypt(
    source: bytes,
    key_crns: List[str],
    aad: Optional[str] = None,
    session: Optional[redstone.Session] = None,
) -> Tuple[bytes, MessageHeader]:
    """Encrypt byte data using a given set of keys from KeyProtect."""

    if session is None:
        session = redstone.get_default_session()

    # generate deks with keyprotect master keys
    data_keys = []
    pt_data_key = b""
    for key_crn in key_crns:
        crn = redstone.crn.loads(key_crn)
        kp = session.service(
            "KeyProtect", region=crn.location, service_instance_id=crn.service_instance
        )

        dek_data = kp.wrap(crn.resource, pt_data_key, aad=[aad])
        if not pt_data_key:
            # plaintext is returned as a utf8 string of base64 in the `plaintext` field
            pt_data_key = base64.b64decode(dek_data["plaintext"].encode("utf-8"))

        # ciphertext is also a utf8 string,
        # but we don't need to do anything but store it for now
        data_keys.append(
            {
                "ciphertext": dek_data["ciphertext"],
                "key_crn": key_crn,
            }
        )

    # we now have all the data keys and plaintext form to do some encryption with

    # bail if we didn't get a 32 byte key from keyprotect,
    # this shouldn't happen but...
    # if it ever does we don't EVER want to go forward with a weak key
    if len(pt_data_key) != 32:
        raise Exception("Plaintext key from KMS was not 256 bits!")

    gcm = AESGCM(pt_data_key)

    # use standard 12 byte nonce
    nonce = os.urandom(
        12
    )  # see: https://cryptography.io/en/latest/random-numbers/#random-number-generation

    # tag is the last 16 bytes
    ciphertext_and_tag = gcm.encrypt(
        nonce, source, aad.encode("utf-8") if aad else None
    )

    # prepend nonce to ct and tag
    encrypted_message = nonce + ciphertext_and_tag

    message_header = MessageHeader(data_keys=data_keys, aad=aad, algorithm="AES256-GCM")

    return message_header._pack() + encrypted_message, message_header


def decrypt(
    source: bytes, session: Optional[redstone.Session] = None
) -> Tuple[bytes, MessageHeader]:
    """Decrypt data previously encrypted with the encrypt function."""

    if session is None:
        session = redstone.get_default_session()

    # unpack headers to get crypto information
    header, message = MessageHeader.from_message_with_body(source)

    pt_data_key = None
    for data_key in header.data_keys:
        # NOTE(mrodden): might be good to prefer a local region here sometime
        crn = redstone.crn.loads(data_key["key_crn"])
        LOG.info("Decrypting data key with master key: %s" % crn)
        kp = session.service(
            "KeyProtect", region=crn.location, service_instance_id=crn.service_instance
        )
        try:
            pt_data_key = kp.unwrap(
                crn.resource, data_key["ciphertext"], aad=[header.aad]
            )
        except Exception as ex:
            LOG.warning("Exception while attempting unwap: %s" % str(ex))
            continue
        if pt_data_key:
            LOG.info("Decrypted data key.")
            break

    if pt_data_key is None:
        raise Exception("Failed to unwrap any keys!")

    # got key, decrypt
    gcm = AESGCM(pt_data_key)
    plaintext_message = gcm.decrypt(
        message[:12], message[12:], header.aad.encode("utf-8") if header.aad else None
    )

    return plaintext_message, header


def get_message_version(message_header: bytes) -> int:
    return int.from_bytes(message_header[:1], byteorder="big")


def _message_version_to_bytes(version: int) -> bytes:
    return version.to_bytes(1, byteorder="big")



# Copyright 2020 Mathew Odden <mathewrodden@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa E501
"""IBM Cloud CRN types and functions.

Parsing a CRN from a (possibly urlencoded) string::

    >>> import redstone.crn
    >>> mycrn = redstone.crn.loads("crn%3Av1%3Astaging%3Apublic%3Aexampleservice%3Aglobal%3Aa%2Fe5d5b304e0f3469f9145bed817f2efe1%3A6fc68009-12f6-4f9e-be69-02cb3b7e0a8d%3A%3A")
    >>> mycrn
    redstone.crn.CRN(prefix='crn', version='v1', cname='staging', ctype='public', service_name='exampleservice', location='global', scope='a/e5d5b304e0f3469f9145bed817f2efe1', service_instance='6fc68009-12f6-4f9e-be69-02cb3b7e0a8d', resource_type='', resource='')
    >>> str(mycrn)
    'crn:v1:staging:public:exampleservice:global:a/e5d5b304e0f3469f9145bed817f2efe1:6fc68009-12f6-4f9e-be69-02cb3b7e0a8d::'
"""

import sys
import urllib.parse


SEGMENTS = [
    "prefix",
    "version",
    "cname",
    "ctype",
    "service_name",
    "location",
    "scope",
    "service_instance",
    "resource_type",
    "resource",
]


class CRN(object):
    def __init__(self, **kwargs):
        for segment in SEGMENTS:
            setattr(self, segment, kwargs.get(segment, ""))

        if not self.prefix:
            self.prefix = "crn"

        if not self.version:
            self.version = "v1"

    def __repr__(self):
        kv_pairs = ["%s=%r" % (seg, getattr(self, seg)) for seg in SEGMENTS]
        return "%s.%s(%s)" % (
            self.__class__.__module__,
            self.__class__.__qualname__,
            ", ".join(kv_pairs),
        )

    def __str__(self):
        return ":".join([getattr(self, seg) for seg in SEGMENTS])


def loads(crn_string):
    if not crn_string.startswith("crn"):
        raise ValueError("Invalid crn prefix: %r" % crn_string)

    if "%3A" in crn_string:
        crn_string = urllib.parse.unquote(crn_string)

    vals = crn_string.split(":")
    if len(vals) != 10:
        raise ValueError("Invalid crn string, must have exactly 10 segments")

    kv_pairs = zip(SEGMENTS, vals)
    return CRN(**dict(kv_pairs))


if __name__ == "__main__":
    print(loads(sys.argv[1]))



# Copyright 2019 Mathew Odden <mathewrodden@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The :mod:`redstone` module is used to interact with various IBM Cloud services.

Basic usage involves getting a :py:class:`Session` object, which can
be used to build a client for interacting with a specific service.

For example, to interact with Resource Controller, we can use the default session to
get a client to talk to Resource Controller::

    >>> import redstone
    >>> session = redstone.get_default_session()
    >>> rc = session.service("ResourceController")

The :py:func:`service` function at the top level can be used as a shortcut for
accessing the default session to build clients. This is equivalent to the above::

    >>> import redstone
    >>> rc = redstone.service("ResourceController")


The default session is constructed lazily on first access and looks for
the environment variable `IBMCLOUD_API_KEY` to use as a credential for
interacting with services.

The default session can be overriden or created manually if desired.
Any clients created using the default session will then use that session
instead.

    >>> import redstone
    >>> redstone.DEFAULT_SESSION = redstone.Session(iam_api_key="...")
    >>> rc = redstone.service("ResourceController")
"""

import os
from typing import Optional  # noqa: F401

from redstone import auth
from redstone import client


DEFAULT_SESSION = None  # Optional[Session]
"""
Holds the current default :py:class:`Session` or
None if no session has been built yet.
"""


class Session(object):
    """
    Session objects are used to create clients used to interact with various services.
    They hold region endpoint information and a credential object that is used
    by the clients for authentication. It's main purpose for sharing and caching
    credentials for use between multiple clients/services.

    A Session can be created manually, but there is also a default session that can
    be accessed by using the :py:func:`get_default_session` function.
    """

    def __init__(self, region=None, iam_api_key=None):
        self.region = region
        self.iam_api_key = iam_api_key
        self.credentials = auth.TokenManager(self.iam_api_key)

    def service(self, service_name, **kwargs):
        client_cls = getattr(client, service_name, None)
        if not client_cls:
            raise ValueError("No client for service '%s'" % service_name)

        if kwargs.get("iam_api_key"):
            del kwargs["iam_api_key"]

        if not kwargs.get("region"):
            kwargs["region"] = self.region

        if not kwargs.get("credentials"):
            kwargs["credentials"] = self.credentials

        return client_cls(**kwargs)


def get_default_session() -> Session:
    """
    Returns the current default session for building clients objects.
    """

    global DEFAULT_SESSION

    if DEFAULT_SESSION is None:
        DEFAULT_SESSION = Session(
            iam_api_key=os.environ.get("IBMCLOUD_API_KEY"),
            region=os.environ.get("IBMCLOUD_REGION"),
        )

    return DEFAULT_SESSION


def service(service_name, **kwargs):
    """Create and return a new client using the :const:`DEFAULT_SESSION`."""
    return get_default_session().service(service_name, **kwargs)



# Copyright 2019 Mathew Odden <mathewrodden@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module holds functionality for authenticating with IBM Cloud services,
which mainly involves getting a token from IAM and using that token
while making requests to the various other services.

Most users will want to use the high level :py:class:`TokenManager`
which provides caching and automatic refresh for tokens.

There is also a low level :py:func:`auth` function that can be used
to request a new token from IAM as needed.
"""

from __future__ import print_function

import base64
import json
import logging
import os
import sys
import threading
import time

import requests


LOG = logging.getLogger(__name__)


class TokenManager(object):
    """
    TokenManager objects are a wrapper around an API key credential, which
    is used to request tokens when they are needed.

    A TokenManager object caches tokens to minimize requests to IAM,
    and also will take care of requesting new tokens when the current
    cached one is expired.

    Example usage::

        tokman = TokenManager(api_key="my-cloud-api-key")

        # get_token() will return a cached version or request one if needed
        iam_token = tokman.get_token()
    """

    def __init__(self, api_key, iam_endpoint=None, use_refresh_token=False):
        self.api_key = api_key
        self.iam_endpoint = iam_endpoint
        self._token_info = {}
        self._lock = threading.RLock()
        self._use_refresh_token = use_refresh_token

    def get_token(self) -> str:
        """
        Retrieve a valid, unexpired token from IAM if needed,
        or return a cached, unexpired token.
        """

        with self._lock:
            if (
                not self._token_info.get("access_token")
                or self.is_refresh_token_expired()
            ):
                self._request_token()
            elif self.is_token_expired():
                if self._use_refresh_token:
                    self._refresh_token()
                else:
                    self._request_token()

            return self._token_info.get("access_token")

    def _request_token(self):
        token_resp = auth(apikey=self.api_key, iam_endpoint=self.iam_endpoint)
        if isinstance(token_resp, dict):
            self._token_info = token_resp
        else:
            raise Exception("Error getting token: %s" % token_resp)

    def _refresh_token(self):
        token_resp = auth(
            refresh_token=self._token_info.get("refresh_token"),
            iam_endpoint=self.iam_endpoint,
        )
        if isinstance(token_resp, dict):
            self._token_info = token_resp
        else:
            raise Exception("Error refreshing token: %s" % token_resp)

    def is_token_expired(self):
        """
        Use to check if the cached IAM token needs to be refreshed.

        Returns:
            bool, True if the token needs to be refreshed, False otherwise
        """
        # refresh even with 20% time still remainig,
        # should be 12 minutes before expiry for 1h tokens
        # and 4 minutes before expiry for 20m tokens
        token_expire_time = self._token_info.get("expiration", 0)
        token_expires_in = self._token_info.get("expires_in", 0)
        return time.time() >= (token_expire_time - (0.2 * token_expires_in))

    def is_refresh_token_expired(self):
        """
        Use to check if the cached IAM Refresh token needs to be refreshed.

        The Refresh token is a different token than the IAM token used
        to interact with services. A Refresh token is a longer lasting token
        that can be used instead of an API key or password credential to request
        a new IAM token. It is useful for some specific cases, where the API key
        or password needs to be dropped and the Refresh token can be used instead
        to generate IAM tokens.

        Returns:
            bool, True if the Refresh token needs to be refreshed, False otherwise
        """
        # no idea how long these last,
        # but some other code suggested up to 30 days,
        # but it was also assuming they expire within 7 days...
        # assume 7 days, because better safe than sorry
        day = 24 * 60 * 60
        refresh_expire_time = self._token_info.get("expiration", 0) + (7 * day)
        return time.time() >= refresh_expire_time


def auth(
    username=None, password=None, apikey=None, refresh_token=None, iam_endpoint=None
):
    """
    Makes a authentication request to the IAM API to retrieve an IAM token and
    IAM Refresh token.

    :param username: Username
    :param password: Password
    :param apikey: IBMCloud/Bluemix API Key
    :param refresh_token: IBM IAM Refresh Token,
        if specified the refresh token is used to authenticate,
        instead of the API key
    :param iam_endpoint: base URL that can be specified
        to override the default IAM endpoint, if one, for example,
        wanted to test against their own IAM or an internal server
    :return: Response
    """
    if not iam_endpoint:
        iam_endpoint = "https://iam.cloud.ibm.com/"

    if iam_endpoint[-1] == "/":
        iam_endpoint = iam_endpoint[:-1]

    api_endpoint = iam_endpoint + "/oidc/token"

    # HTTP Headers
    headers = {"Authorization": "Basic Yng6Yng=", "Accept": "application/json"}

    # HTTP Payload
    data = {
        "response_type": "cloud_iam",
        "uaa_client_id": "cf",
        "uaa_client_secret": "",
    }

    # Setup grant type
    if apikey:
        data["grant_type"] = "urn:ibm:params:oauth:grant-type:apikey"
        data["apikey"] = apikey
    elif refresh_token:
        data["grant_type"] = "refresh_token"
        data["refresh_token"] = refresh_token
    elif username and password:
        data["grant_type"] = "password"
        data["username"] = username
        data["password"] = password
    else:
        raise ValueError(
            "Must specify one of username/password, apikey, or refresh_token!"
        )

    resp = requests.post(api_endpoint, data=data, headers=headers)

    if resp.status_code == 200:
        return resp.json()

    return resp.text


def get_orgs(bearer_token):
    api_endpoint = "https://api.ng.bluemix.net/v2/organizations"

    headers = {
        "Content-Type": "application/x-www-form-urlencoded;charset=utf",
        "Authorization": "Bearer %s" % bearer_token,
        "Accept": "application/json;charset=utf-8",
    }

    resp = requests.get(api_endpoint, headers=headers)
    return resp.text


def get_spaces(bearer_token, spaces_path):
    api_endpoint = "https://api.ng.bluemix.net%s" % spaces_path

    headers = {
        "Content-Type": "application/x-www-form-urlencoded;charset=utf",
        "Authorization": "Bearer %s" % bearer_token,
        "Accept": "application/json;charset=utf-8",
    }

    resp = requests.get(api_endpoint, headers=headers)
    return resp.text


def find_space_and_org(bearer_token, org_name, space_name):
    org_resp = get_orgs(bearer_token)
    org_data = json.loads(org_resp)

    for org in org_data["resources"]:
        if org_name == org.get("entity", {}).get("name"):
            org_info = org
            break

    space_resp = get_spaces(bearer_token, org_info["entity"]["spaces_url"])
    space_data = json.loads(space_resp)

    for space in space_data["resources"]:
        if space_name == space.get("entity", {}).get("name"):
            space_info = space
            break

    return org_info, space_info


def inspect_token(token):
    parts = token.split(".")[:2]
    decoded_parts = []
    for part in parts:
        padding = "=" * (len(part) % 4)
        part = str(part)
        decoded_part = base64.urlsafe_b64decode(part + padding)
        try:
            decoded_part = json.loads(decoded_part.decode("utf8"))
        except ValueError:
            pass
        decoded_parts.append(decoded_part)

    return decoded_parts


def main():
    api_key = None

    # iterate through possible things that could be used to get us the key,
    # last one that is not None or empty will win
    possible_keys = [
        os.environ.get("BLUEMIX_API_KEY"),
        os.environ.get("IBMCLOUD_API_KEY"),
    ]
    for pkey in possible_keys:
        if pkey:
            api_key = pkey

    if not api_key:
        print("error: please set BLUEMIX_API_KEY or IBMCLOUD_API_KEY", file=sys.stderr)
        return 1

    tokman = TokenManager(api_key=api_key)
    print(tokman.get_token())


if __name__ == "__main__":
    sys.exit(main())



"""Command-line tool to encrypt/decrypt data using IBM KeyProtect

This is meant to be a quick way to encrypt/decrypt via the command line
with shell or other tools. It only works on data sizes up to about 2 gigabytes,
and will not be the most performant way.

It also does not support some more advanced features like reading the encrypted
message metadata header, or providing additional authentication data during encryption.
If you need these features please use the 'crypto' module directly.
"""

import argparse
import base64
import os
import sys

from redstone import crypto


key_crn_env_var = "RSCRYPTO_KEY_CRNS"


def main():
    p = argparse.ArgumentParser()

    subp = p.add_subparsers(dest="action")

    ep = subp.add_parser("encrypt")
    ep.add_argument(
        "--key-crns",
        help=(
            "Space separated list of CRNs of KMS Keys to use during encryption. "
            "At least one CRN must be specified here or by setting %s" % key_crn_env_var
        ),
    )
    ep.add_argument(
        "infile",
        type=argparse.FileType("rb"),
        help=(
            "Input file to encrypt. Output will be written to stdout. "
            "Use '-' to read from stdin."
        ),
    )

    dp = subp.add_parser("decrypt")
    dp.add_argument(
        "infile",
        type=argparse.FileType("rb"),
        help=(
            "Input file to decrypt. Output will be written to stdout. "
            "Use '-' to read from stdin."
        ),
    )

    args = p.parse_args()

    api_key = os.getenv("IBMCLOUD_API_KEY")
    if not api_key:
        print("'IBMCLOUD_API_KEY' must be set", file=sys.stderr)
        return 1

    if args.action == "encrypt":
        key_crns = args.key_crns or os.getenv(key_crn_env_var)
        if not key_crns:
            print(
                (
                    "No key CRNs specified. "
                    "Please pass the CRNs with --key-crns or set '%s'"
                )
                % key_crn_env_var,
                file=sys.stderr,
            )
            return 1
        key_crns = key_crns.split(" ")

        data = args.infile.read()
        if isinstance(data, str):
            data = data.encode("utf-8")

        message, _ = crypto.encrypt(
            data, key_crns=key_crns, aad="redstone.crypto.__main__"
        )
        print(base64.b64encode(message).decode("utf-8"), file=sys.stdout)
    elif args.action == "decrypt":
        message, _ = crypto.decrypt(base64.b64decode(args.infile.read()))
        # write raw bytes, as we don't know the encoding of the data
        sys.stdout.buffer.write(message)
    else:
        p.print_usage(file=sys.stderr)
        return 1


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2021 Mathew Odden
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf8") as fh:
    long_desc = fh.read()

setup(
    name="redstone",
    version="0.6.0",
    author="Mathew Odden",
    author_email="mathewrodden@gmail.com",
    url="https://github.com/IBM/redstone",
    description="A Pythonic IBM Cloud SDK",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=["requests[security]", "cryptography"],
    extras_require={
        "docs": ["sphinx>=3.1", "sphinx_rtd_theme"],
    },
    entry_points={
        "console_scripts": [
            "rs-crypto = redstone.crypto.__main__:main",
            "rs-keyprotect = redstone.keyprotect.cli:main",
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.6",
)



import logging
import unittest

import redstone


TEST_INSTANCE_NAME_PREFIX = "my-test-instance"


class ResourceControllerTestCase(unittest.TestCase):
    def tearDown(self):
        self.cleanup_instances()

    def cleanup_instances(self):
        rc = redstone.service("ResourceController")

        to_delete = filter(
            lambda x: x.get("name").startswith(TEST_INSTANCE_NAME_PREFIX),
            rc.list_instances(),
        )
        for inst in to_delete:
            rc.delete_instance(inst.get("id"))

    def test_delete_with_crn(self):
        rc = redstone.service("ResourceController")

        inst_id, crn = rc.create_instance(
            name=TEST_INSTANCE_NAME_PREFIX,
            plan_id=rc.KEYPROTECT_PLAN_ID,
            region="us-south",
        )

        rc.delete_instance(crn)

        self.assertIsNone(rc.get_instance(crn))

    def test_delete_with_id(self):
        rc = redstone.service("ResourceController")

        inst_id, crn = rc.create_instance(
            name=TEST_INSTANCE_NAME_PREFIX,
            plan_id=rc.KEYPROTECT_PLAN_ID,
            region="us-south",
        )

        rc.delete_instance(inst_id)

        self.assertIsNone(rc.get_instance(crn))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    unittest.main()
